@@ -6,6 +6,7 @@ Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-27%3A-
 # Distributed under the terms of the Modified BSD License.
 import json
 from http import HTTPStatus
+from typing import Any, Dict, List
 
 try:
     from jupyter_client.jsonutil import json_default
@@ -15,19 +16,42 @@ except ImportError:
 from jupyter_core.utils import ensure_async
 from tornado import web
 
-from jupyter_server.auth.decorator import authorized
+from jupyter_server.auth.decorator import allow_unauthenticated, authorized
 from jupyter_server.base.handlers import APIHandler, JupyterHandler, path_regex
 from jupyter_server.utils import url_escape, url_path_join
 
 AUTH_RESOURCE = "contents"
 
 
-def validate_model(model, expect_content):
+def _validate_keys(expect_defined: bool, model: Dict[str, Any], keys: List[str]):
+    """
+    Validate that the keys are defined (i.e. not None) or not (i.e. None)
+    """
+
+    if expect_defined:
+        errors = [key for key in keys if model[key] is None]
+        if errors:
+            raise web.HTTPError(
+                500,
+                f"Keys unexpectedly None: {errors}",
+            )
+    else:
+        errors = {key: model[key] for key in keys if model[key] is not None}  # type: ignore[assignment]
+        if errors:
+            raise web.HTTPError(
+                500,
+                f"Keys unexpectedly not None: {errors}",
+            )
+
+
+def validate_model(model, expect_content=False, expect_hash=False):
     """
     Validate a model returned by a ContentsManager method.
 
     If expect_content is True, then we expect non-null entries for 'content'
     and 'format'.
+
+    If expect_hash is True, then we expect non-null entries for 'hash' and 'hash_algorithm'.
     """
     required_keys = {
         "name",
@@ -40,6 +64,8 @@ def validate_model(model, expect_content):
         "content",
         "format",
     }
+    if expect_hash:
+        required_keys.update(["hash", "hash_algorithm"])
     missing = required_keys - set(model.keys())
     if missing:
         raise web.HTTPError(
@@ -47,21 +73,10 @@ def validate_model(model, expect_content):
             f"Missing Model Keys: {missing}",
         )
 
-    maybe_none_keys = ["content", "format"]
-    if expect_content:
-        errors = [key for key in maybe_none_keys if model[key] is None]
-        if errors:
-            raise web.HTTPError(
-                500,
-                f"Keys unexpectedly None: {errors}",
-            )
-    else:
-        errors = {key: model[key] for key in maybe_none_keys if model[key] is not None}  # type: ignore[assignment]
-        if errors:
-            raise web.HTTPError(
-                500,
-                f"Keys unexpectedly not None: {errors}",
-            )
+    content_keys = ["content", "format"]
+    _validate_keys(expect_content, model, content_keys)
+    if expect_hash:
+        _validate_keys(expect_hash, model, ["hash", "hash_algorithm"])
 
 
 class ContentsAPIHandler(APIHandler):
@@ -122,20 +137,41 @@ class ContentsHandler(ContentsAPIHandler):
             raise web.HTTPError(400, "Content %r is invalid" % content_str)
         content = int(content_str or "")
 
+        hash_str = self.get_query_argument("hash", default="0")
+        if hash_str not in {"0", "1"}:
+            raise web.HTTPError(400, f"Content {hash_str!r} is invalid")
+        require_hash = int(hash_str)
+
         if not cm.allow_hidden and await ensure_async(cm.is_hidden(path)):
             await self._finish_error(
                 HTTPStatus.NOT_FOUND, f"file or directory {path!r} does not exist"
             )
+
         try:
-            model = await ensure_async(
-                self.contents_manager.get(
-                    path=path,
-                    type=type,
-                    format=format,
-                    content=content,
+            expect_hash = require_hash
+            try:
+                model = await ensure_async(
+                    self.contents_manager.get(
+                        path=path,
+                        type=type,
+                        format=format,
+                        content=content,
+                        require_hash=require_hash,
+                    )
                 )
-            )
-            validate_model(model, expect_content=content)
+            except TypeError:
+                # Fallback for ContentsManager not handling the require_hash argument
+                # introduced in 2.11
+                expect_hash = False
+                model = await ensure_async(
+                    self.contents_manager.get(
+                        path=path,
+                        type=type,
+                        format=format,
+                        content=content,
+                    )
+                )
+            validate_model(model, expect_content=content, expect_hash=expect_hash)
             self._finish_model(model, location=False)
         except web.HTTPError as exc:
             # 404 is okay in this context, catch exception and return 404 code to prevent stack trace on client
@@ -165,7 +201,7 @@ class ContentsHandler(ContentsAPIHandler):
             raise web.HTTPError(400, f"Cannot rename file or directory {path!r}")
 
         model = await ensure_async(cm.update(model, path))
-        validate_model(model, expect_content=False)
+        validate_model(model)
         self._finish_model(model)
 
     async def _copy(self, copy_from, copy_to=None):
@@ -178,7 +214,7 @@ class ContentsHandler(ContentsAPIHandler):
         )
         model = await ensure_async(self.contents_manager.copy(copy_from, copy_to))
         self.set_status(201)
-        validate_model(model, expect_content=False)
+        validate_model(model)
         self._finish_model(model)
 
     async def _upload(self, model, path):
@@ -186,7 +222,7 @@ class ContentsHandler(ContentsAPIHandler):
         self.log.info("Uploading file to %s", path)
         model = await ensure_async(self.contents_manager.new(model, path))
         self.set_status(201)
-        validate_model(model, expect_content=False)
+        validate_model(model)
         self._finish_model(model)
 
     async def _new_untitled(self, path, type="", ext=""):
@@ -196,7 +232,7 @@ class ContentsHandler(ContentsAPIHandler):
             self.contents_manager.new_untitled(path=path, type=type, ext=ext)
         )
         self.set_status(201)
-        validate_model(model, expect_content=False)
+        validate_model(model)
         self._finish_model(model)
 
     async def _save(self, model, path):
@@ -205,7 +241,7 @@ class ContentsHandler(ContentsAPIHandler):
         if not chunk or chunk == -1:  # Avoid tedious log information
             self.log.info("Saving file at %s", path)
         model = await ensure_async(self.contents_manager.save(model, path))
-        validate_model(model, expect_content=False)
+        validate_model(model)
         self._finish_model(model)
 
     @web.authenticated
@@ -356,8 +392,15 @@ class ModifyCheckpointsHandler(ContentsAPIHandler):
 class NotebooksRedirectHandler(JupyterHandler):
     """Redirect /api/notebooks to /api/contents"""
 
-    SUPPORTED_METHODS = ("GET", "PUT", "PATCH", "POST", "DELETE")  # type:ignore[assignment]
+    SUPPORTED_METHODS = (
+        "GET",
+        "PUT",
+        "PATCH",
+        "POST",
+        "DELETE",
+    )  # type:ignore[assignment]
 
+    @allow_unauthenticated
     def get(self, path):
         """Handle a notebooks redirect."""
         self.log.warning("/api/notebooks is deprecated, use /api/contents")
