@@ -154,7 +154,7 @@ class ZMQChannelsWebsocketConnection(BaseKernelWebsocketConnection):
             self.channels[channel] = stream = meth(identity=identity)
             stream.channel = channel
 
-    def nudge(self):  # noqa
+    def nudge(self):
         """Nudge the zmq connections with kernel_info_requests
         Returns a Future that will resolve when we have received
         a shell or control reply and at least one iopub message,
@@ -376,7 +376,7 @@ class ZMQChannelsWebsocketConnection(BaseKernelWebsocketConnection):
                     if not stream.closed():
                         stream.close()
                 self.disconnect()
-                return
+                return None
 
         self.multi_kernel_manager.add_restart_callback(self.kernel_id, self.on_kernel_restarted)
         self.multi_kernel_manager.add_restart_callback(
@@ -438,7 +438,7 @@ class ZMQChannelsWebsocketConnection(BaseKernelWebsocketConnection):
         try:
             ZMQChannelsWebsocketConnection._open_sockets.remove(self)
             self._close_future.set_result(None)
-        except Exception:  # noqa
+        except Exception:
             pass
 
     def handle_incoming_message(self, incoming_msg: str) -> None:
